from typing import Optional

import requests

from spoffy.models import Token
from spoffy.client.base import SyncClient, ClientCommon
from spoffy.sansio import Request, Response
from spoffy.spotify import SyncSpotify


class RequestsClient(SyncClient):
    """
    Client implementation using requests as a http backend

    :param session: A :class:`~requests.Session` object
    """

    def __init__(
        self,
        *,
        session: Optional[requests.Session] = None,
        access_token: Optional[str] = None,
        token: Optional[Token] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        redirect_uri: Optional[str] = None,
        scope: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:

        self.session: requests.Session = session or requests  # type: ignore
        super().__init__(
            access_token,
            token,
            client_id,
            client_secret,
            redirect_uri,
            scope,
            state,
        )

    def request(self, request: Request) -> Response:
        resp = self.session.request(
            method=request.method,
            url=request.url,
            data=request.body,
            headers=request.headers,
        )
        response = Response(
            request, resp.status_code, resp.headers, resp.content
        )
        response.raise_for_status()
        return response


def make_spotify(
    *,
    session: Optional[requests.Session] = None,
    access_token: Optional[str] = None,
    token: Optional[Token] = None,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    redirect_uri: Optional[str] = None,
    scope: Optional[str] = None,
    state: Optional[str] = None,
) -> SyncSpotify:
    """
    Convenience factory to build
    Spotify API wrapper using the requests http library.
    Accepts all arguments of :class:`~RequestsClient`
    """
    return SyncSpotify(
        RequestsClient(
            session=session,
            access_token=access_token,
            token=token,
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            scope=scope,
            state=state,
        )
    )


RequestsClient.__doc__ += ClientCommon.__doc__.split("----")[1]  # type: ignore
make_spotify.__doc__ += ClientCommon.__doc__.split("----")[1]  # type: ignore



from pydantic import BaseModel, BaseConfig, Extra


class Cfg(BaseConfig):
    extra = Extra.allow


class CustomMeta(BaseModel.__class__):  # type:ignore
    def __new__(mcs, name, bases, namespace):
        ret = super().__new__(mcs, name, bases, namespace)
        for field in ret.__fields__.values():
            field.required = False
            field.prepare()
        return ret


class SpotifyObject(BaseModel, metaclass=CustomMeta):
    Config = Cfg

    def __init__(self, **data) -> None:
        super().__init__(**data)
        for field in self.__fields_set__:
            if field not in self.__fields__:
                value = getattr(self, field)
                if isinstance(value, dict):
                    setattr(self, field, SpotifyObject(**value))
                elif (
                    value
                    and isinstance(value, list)
                    and isinstance(value[0], dict)
                ):
                    setattr(self, field, [SpotifyObject(**v) for v in value])



from typing import List, Optional as Opt

from spoffy.models.image import Image
from spoffy.models.users import UserBase
from spoffy.models.core import Track, TrackLink, ExternalUrls, Followers
from spoffy.models.paging import OffsetPaging
from spoffy.models.base import SpotifyObject


class PlaylistSnapshotId(SpotifyObject):
    snapshot_id: str


class PlaylistOwner(UserBase):
    display_name: str


class PlaylistTrackAddedBy(UserBase):
    pass


class VideoThumbnail(SpotifyObject):
    url: Opt[str]


class PlaylistTrackTrack(Track):
    available_markets: Opt[List[str]] = None
    linked_from: Opt[TrackLink] = None
    is_playable: Opt[bool] = None


class PlaylistTrack(SpotifyObject):
    added_at: Opt[str]  # todo: datetime
    added_by: Opt[PlaylistTrackAddedBy]
    is_local: bool
    track: PlaylistTrackTrack


class PlaylistTracksPaging(OffsetPaging):
    items: List[PlaylistTrack]


class PlaylistBase(SpotifyObject):
    """
    :param collaborative: Whether playlist is collaborative
    :type collaborative: :class:`bool`
    :param external_urls: Extern urls for playlist
    :type external_urls: :class:`ExternalUrls`
    :param href: URL to this object
    :type href: :class:`str`
    :param id: Playlist ID
    :type id: :class:`str`
    :param images: Playlist images
    :type images: :class:`list[]` of :class:`Image` objects
    :param name: The playlist name
    :type name: :class:`str`
    :param owner: The playlist owner
    :type owner: :class:`PlaylistOwner`
    :param public: Whether playlist is public
    :type public: :class:`bool`
    :param snapshot_id: The playlist snapshot ID
    :type snapshot_id: :class:`str`
    :param type: The type of object ('playlist')
    :type type: :class:`str`
    :param uri: The playlist URI
    :type uri: :class:`str`
    """

    collaborative: bool
    external_urls: ExternalUrls
    href: str
    id: str
    images: List[Image]
    name: str
    owner: PlaylistOwner
    public: bool
    snapshot_id: str
    type: str
    uri: str


class TracksHref(SpotifyObject):
    """
    """

    href: str
    total: int


class PlaylistSimple(PlaylistBase):
    """
    A simplified playlist object

    :param tracks: Info to access playlist tracks
    :type tracks: :class:`TracksHref`
    """

    tracks: TracksHref


PlaylistSimple.__doc__ += PlaylistBase.__doc__  # type: ignore


class Playlist(PlaylistBase):
    """
    A complete playlist object

    :param description: Text description of playlist
    :type description: :class:`str`
    :param followers: Info about playlist followers
    :type followers: :class:`Followers`
    :param tracks: Playlist tracks paginated
    :type tracks: :class:`PlaylistTracksPaging`
    """

    description: Opt[str]
    followers: Followers
    tracks: PlaylistTracksPaging


class PlaylistSimplePaging(OffsetPaging):
    items: List[PlaylistSimple]


Playlist.__doc__ += PlaylistBase.__doc__  # type: ignore



from urlobject import URLObject


class AuthMixin:
    def get_authorize_url(self, **kwargs) -> str:
        params = dict(
            client_id=self.client.client_id,  # type: ignore
            redirect_uri=self.client.redirect_uri,  # type: ignore
            scope=self.client.scope,  # type: ignore
            state=self.client.state,  # type: ignore
            response_type="code",
        )
        params.update(kwargs)
        return str(
            URLObject(
                self.client.authorize_url  # type: ignore
            ).add_query_params(**params)
        )



from spoffy.models.base import SpotifyObject


class AudioFeatures(SpotifyObject):
    duration_ms: int
    key: int
    mode: int
    time_signature: int
    acousticness: float
    danceability: float
    energy: float
    instrumentalness: float
    liveness: float
    loudness: float
    speechiness: float
    valence: float
    tempo: float
    id: str
    uri: str
    track_href: str
    analysis_url: str
    type: str



from abc import abstractmethod


from spoffy.client.common import ClientCommon
from spoffy.sansio import Response
from spoffy.sansio import Request


class AsyncClient(ClientCommon):
    """
    Base class for async spotify client.
    Subclasses must implement the :meth:`~AsyncClient.request` method
    """

    @abstractmethod
    async def request(self, request: Request) -> Response:
        pass


class SyncClient(ClientCommon):
    """
    Base class for a sync spotify client.
    Subclasses must implement the :meth:`~SyncClient.request` method
    """

    @abstractmethod
    def request(self, request: Request) -> Response:
        pass


SyncClient.__doc__ += ClientCommon.__doc__.split("----")[1]  # type: ignore
AsyncClient.__doc__ += ClientCommon.__doc__.split("----")[1]  # type: ignore



from typing import Optional, Union, MutableMapping

from spoffy.models import Token
from spoffy.sansio import Response
from spoffy.sansio import Request


class ClientCommon:
    """
    A sansio client implementation.
    This does no IO on its own.

    ----

    :param access_token: A Spotify access token
        Supersedes the `token` parameter
    :param token: A token object, either for user or client
    :param client_id: Client ID for oauth login
    :param client_secret: Client secret for oauth login
    :param redirect_uri: Redirect URI for oauth login
    :param scope: Space separated list of scopes for oauth login
    :param state: State string for oauth login
    """

    _default_prefix = "https://api.spotify.com/v1"

    #: Base authorize url
    authorize_url = "https://accounts.spotify.com/authorize"
    #: Token url for oauth and refresh
    token_url = "https://accounts.spotify.com/api/token"
    #: URL prefix used for all Spotify API calls
    #: This can be changed if using a proxy
    base_url = _default_prefix

    def __init__(
        self,
        access_token: Optional[str] = None,
        token: Optional[Token] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        redirect_uri: Optional[str] = None,
        scope: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        self._access_token = access_token
        self.token = token
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.scope = scope
        self.state = state

    def prepare_request(
        self,
        method: str,
        url: str,
        params: Optional[MutableMapping] = None,
        body: Optional[Union[bytes, MutableMapping]] = None,
        headers: Optional[MutableMapping[str, str]] = None,
        access_token: Optional[str] = None,
    ) -> Request:
        """
        Create a request object from the given arguments

        :param method: The request http method
            ( ``GET`` / ``POST`` / ``PUT`` / ``DELETE`` )
        :param url: The request URL, relative URLs get prefixed with
            :py:attr:`~base_url`
        :param params: Dict of query parameters to add to URL
        :param body: Request body as either bytes or json serializable dict
        :param headers: Dict of headers to add to request
        :param access_token: Override client access token
            If not provided and no explicit ``Authorization`` header,
            the :py:attr:`~access_token` stored on this instance is used
        """
        if (
            url.startswith(self._default_prefix)
            and self._default_prefix != self.base_url
        ):
            url = url.replace(self._default_prefix, self.base_url)
        elif url.startswith("http://") or url.startswith("https://"):
            pass
        else:
            url = self.base_url + url
        headers = headers or {}
        if not access_token and "Authorization" not in headers:
            access_token = self.access_token
        return Request(
            method=method,
            url=url,
            params=params,
            body=body,
            headers=headers,
            access_token=access_token,
        )

    @property
    def access_token(self) -> Union[str, None]:
        """
        Property to get the current access token
        Returns either the access_token passed to the constructor
        or the access token from this client's :py:attr:`~token` object.
        Returns ``None`` if no token
        """
        if self._access_token:
            return self._access_token
        elif self.token:
            return self.token.access_token
        return None

    @access_token.setter
    def access_token(self, access_token: str):
        self.access_token = access_token

    def load(self, response: Response, target):
        """
        Load a response data to a ``SpotifyObject`` object

        :param response: The response to load
        :param target: The target class
        """
        if target is None:
            return None
        data = response.json
        if data is None:
            return None
        return target(**data)



from base64 import b64encode
from typing import Sequence, Optional, Dict, Union, Any

from spoffy.sansio import Request
from spoffy.client.common import ClientCommon
from spoffy import models


def returns(type_):
    def decorator(func):
        func.__spotify_object__ = type_
        return func

    return decorator


class RequestBuilder:
    def __init__(self, client: ClientCommon):
        self.client = client

    def b(self, *args, **kwargs):
        return self.client.prepare_request(*args, **kwargs)


class Artists(RequestBuilder):
    @returns(models.Artist)
    def artist(self, artist_id: str) -> Request:
        """
        Get an artist by Spotify ID

        :param artist_id:
        """
        return self.b(method="GET", url="/artists/{}".format(artist_id))

    @returns(models.ArtistsCollection)
    def many_artists(self, ids: Sequence[str]) -> Request:
        """
        Get several artists by ID

        :param ids: List of Spotify artist IDs
        """
        return self.b("GET", "/artists", params=dict(ids=",".join(ids)))

    @returns(models.AlbumSimplePaging)
    def artist_albums(
        self,
        artist_id: str,
        include_groups: Sequence[str] = None,
        market: str = None,
    ) -> Request:
        """
        Get all artist's albums

        :param artist_id:
        :param include_groups:
        :param market:
        """
        params = {}
        if include_groups:
            params["include_groups"] = ",".join(include_groups)
        if market is not None:
            params["market"] = market
        return self.b(
            method="GET",
            url="/artists/{}/albums".format(artist_id),
            params=params,
        )

    @returns(models.RelatedArtistsCollection)
    def related(self, artist_id: str) -> Request:
        """
        Get related artists for given artist_id
        """
        return self.b("GET", "/artists/{}/related-artists".format(artist_id))


class Playlists(RequestBuilder):
    @returns(models.Playlist)
    def playlist(
        self,
        playlist_id: str,
        market: Optional[str] = None,
        fields: Optional[str] = None,
    ) -> Request:
        return self.b(
            "GET",
            "/playlists/{}".format(playlist_id),
            params=_clear_nones(dict(market=market, fields=fields)),
        )

    @returns(models.PlaylistTracksPaging)
    def playlist_tracks(
        self,
        playlist_id: str,
        market: Optional[str] = None,
        fields: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Request:
        return self.b(
            "GET",
            "/playlists/{}/tracks".format(playlist_id),
            params=_clear_nones(
                dict(market=market, fields=fields, limit=limit, offset=offset)
            ),
        )

    @returns(models.Playlist)
    def create_playlist(
        self,
        user_id: str,
        name: str,
        public: Optional[bool] = None,
        collaborative: Optional[bool] = None,
        description: Optional[str] = None,
    ) -> Request:
        """
        Create a playlist on a user's account

        :param user_id: ID of the user account
        :param name: The playlist name
        :param public: Whether the playlist should be public
        :param collaborative: Whether the playlist should be collaborative
        :param description: Playlist description, shown in Spotify clients
        """
        return self.b(
            "POST",
            "/users/{}/playlists".format(user_id),
            body=_clear_nones(
                dict(
                    name=name,
                    public=public,
                    collaborative=collaborative,
                    description=description,
                )
            ),
        )

    @returns(models.PlaylistSnapshotId)
    def add_tracks_to_playlist(
        self,
        playlist_id: str,
        uris: Sequence[str],
        position: Optional[int] = None,
    ) -> Request:
        return self.b(
            "POST",
            "/playlists/{}/tracks".format(playlist_id),
            body=_clear_nones(dict(uris=uris, position=position)),
        )

    @returns(models.PlaylistSimplePaging)
    def my_playlists(
        self, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> Request:
        return self.b(
            "GET",
            "/me/playlists",
            params=_clear_nones(dict(limit=limit, offset=offset)),
        )


class Albums(RequestBuilder):
    """
    Interface to the Spotify albums API

    https://developer.spotify.com/documentation/web-api/reference/albums/
    """

    @returns(models.Album)
    def album(self, album_id: str, market: Optional[str] = None) -> Request:
        """
        Get a single album

        :param album_id: The Spotify album ID to fetch
        :param market: ISO2A country to enable relinking
        """
        params = dict(market=market) if market is not None else None
        return self.b("GET", "/albums/{}".format(album_id), params=params)

    @returns(models.AlbumTracksPaging)
    def album_tracks(
        self,
        album_id: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Request:
        """
        Get tracks from the given album

        :param album_id: The Spotify album ID to fetch tracks from
        :param market: ISO2A country to enable relinking
        :param limit: Max number of results to retrieve
        :param offset: Pagination offset
        """
        params = _clear_nones(dict(market=market, limit=limit, offset=offset))
        return self.b(
            "GET", "/albums/{}/tracks".format(album_id), params=params
        )


class Tracks(RequestBuilder):
    """
    Interface to Spotify tracks API

    https://developer.spotify.com/documentation/web-api/reference/tracks
    """

    @returns(models.Track)
    def track(self, track_id: str, market: Optional[str] = None) -> Request:
        """
        Get a single track

        :param track_id: The Spotify track ID to fetch
        :param market: ISO2A country to enable track relinking
        """
        params = {"market": market} if market is not None else None
        return self.b("GET", "/tracks/{}".format(track_id), params=params)

    @returns(models.TracksCollection)
    def many_tracks(
        self, track_ids: Sequence[str], market: Optional[str] = None
    ) -> Request:
        """
        Get several tracks by ID

        :param track_ids: List of up to 50 track IDs
        :param market: ISO2A country to enable track relinking
        """
        return self.b(
            "GET",
            "/tracks",
            params=_clear_nones(dict(ids=",".join(track_ids), market=market)),
        )

    @returns(models.AudioFeatures)
    def audio_features(self, track_id: str) -> Request:
        """
        Get audio features for a single track

        :param track_id: The Spotify track ID to fetch audio features for
        """
        return self.b("GET", "/audio-features/{}".format(track_id))

    @returns(models.AudioFeaturesCollection)
    def many_audio_features(self, track_ids: Sequence[str]) -> Request:
        """
        Get audio features for several tracks

        :param track_ids: List of up to 50 track IDs
        """
        return self.b(
            "GET", "/audio-features", params={"ids": ",".join(track_ids)}
        )


class Users(RequestBuilder):
    @returns(models.UserPublic)
    def user(self, user_id: str) -> Request:
        return self.b("GET", "/users/{}".format(user_id))

    @returns(models.UserPrivate)
    def me(self) -> Request:
        return self.b("GET", "/me")


class Player(RequestBuilder):
    @returns(None)
    def play(
        self,
        device_id: Optional[str] = None,
        context_uri: Optional[str] = None,
        uris: Optional[Sequence[str]] = None,
        offset: Optional[Dict[str, Union[str, int]]] = None,
        position_ms: Optional[int] = None,
    ) -> Request:
        """
        Play an album, track or playlist on Spotify
        """
        body = _clear_nones(
            dict(
                context_uri=context_uri,
                uris=uris,
                offset=offset,
                position_ms=position_ms,
            )
        )
        params = dict(device_id=device_id) if device_id is not None else None
        return self.b("PUT", "/me/player/play", body=body, params=params)

    @returns(models.CurrentPlayback)  # TODO: This returns None sometimes
    def current_playback(self, market: Optional[str] = None) -> Request:
        """
        Get current user's current playback status
        """
        return self.b(
            "GET", "/me/player", params=_clear_nones(dict(market=market))
        )

    @returns(None)
    def next_track(self, device_id: Optional[str] = None) -> Request:
        """
        Skip to next track in Spotify player
        """
        params = dict(device_id=device_id) if device_id is not None else None
        return self.b("PUT", "/me/player/next", params=params)

    @returns(None)
    def previous_track(self, device_id: Optional[str] = None) -> Request:
        """
        Skip to previous track in Spotify player
        """
        params = dict(device_id=device_id) if device_id is not None else None
        return self.b("PUT", "/me/player/previous", params=params)

    @returns(None)
    def pause(self, device_id: Optional[str] = None) -> Request:
        params = dict(device_id=device_id) if device_id is not None else None
        return self.b("PUT", "/me/player/pause", params=params)

    @returns(None)
    def set_volume(
        self, volume_percent: int, device_id: Optional[str] = None
    ) -> Request:
        params = _clear_nones(
            dict(device_id=device_id, volume_percent=volume_percent)
        )
        return self.b("PUT", "/me/player/volume", params=params)

    @returns(models.PlayHistoryPaging)
    def recently_played(
        self,
        limit: int = 20,
        before: Optional[int] = None,
        after: Optional[int] = None,
    ) -> Request:
        params = _clear_nones(dict(limit=limit, before=before, after=after))
        return self.b("GET", "/me/player/recently-played", params=params)

    @returns(models.DevicesCollection)
    def devices(self):
        """
        Get user's available playback devices
        """
        return self.b("GET", "/me/player/devices")


class Search(RequestBuilder):
    """
    Interface to the Spotify search API

    https://developer.spotify.com/documentation/web-api/reference/search/search/
    """

    @returns(models.SearchResults)
    def search(
        self,
        q: str,
        type: Sequence[str],
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        include_external: Optional[str] = None,
    ) -> Request:
        params = _clear_nones(
            dict(
                q=q,
                type=",".join(type),
                market=market,
                limit=limit,
                offset=offset,
                include_external=include_external,
            )
        )
        return self.b("GET", "/search", params=params)

    @returns(models.SearchResults)
    def artists(
        self,
        q: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        include_external: Optional[str] = None,
    ) -> Request:
        """
        Convenience method to search for artists.
        Equivalent to `search(q, types='artist', ...)`

        :param q: The search query
        """
        return self.search(
            q=q,
            type=["artist"],
            market=market,
            limit=limit,
            offset=offset,
            include_external=include_external,
        )

    @returns(models.SearchResults)
    def albums(
        self,
        q: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        include_external: Optional[str] = None,
    ) -> Request:
        """
        Convenience method to search for albums.
        Equivalent to `search(q, types='album', ...)`

        :param q: The search query
        """
        return self.search(
            q=q,
            type=["album"],
            market=market,
            limit=limit,
            offset=offset,
            include_external=include_external,
        )

    @returns(models.SearchResults)
    def tracks(
        self,
        q: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        include_external: Optional[str] = None,
    ) -> Request:
        """
        Convenience method to search for tracks.
        Equivalent to `search(q, types='track', ...)`

        :param q: The search query
        """
        return self.search(
            q=q,
            type=["track"],
            market=market,
            limit=limit,
            offset=offset,
            include_external=include_external,
        )

    @returns(models.SearchResults)
    def playlists(
        self,
        q: str,
        market: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        include_external: Optional[str] = None,
    ) -> Request:
        """
        Convenience method to search for playlists.
        Equivalent to `search(q, types='playlist', ...)`

        :param q: The search query
        """
        return self.search(
            q=q,
            type=["playlist"],
            market=market,
            limit=limit,
            offset=offset,
            include_external=include_external,
        )


class Library(RequestBuilder):
    """
    Manage user's saved tracks and albums

    https://developer.spotify.com/documentation/web-api/reference/library/
    """

    @returns(models.TopArtistsPaging)
    def top_artists(
        self,
        limit: Optional[int] = 20,
        offset: Optional[int] = None,
        time_range: Optional[str] = None,
    ) -> Request:
        """
        Get current user's top artists

        :param limit: Max number of results
        :param offset: Pagination offset
        :time_range: Top artists over specified time period
        """
        return self.b(
            "GET",
            "/me/top/artists",
            params=_clear_nones(
                dict(limit=limit, offset=offset, time_range=time_range)
            ),
        )

    @returns(models.TopTracksPaging)
    def top_tracks(
        self,
        limit: Optional[int] = 20,
        offset: Optional[int] = None,
        time_range: Optional[str] = None,
    ) -> Request:
        """
        Get current user's top artists

        :param limit: Max number of results
        :param offset: Pagination offset
        :time_range: Top artists over specified time period
        """
        return self.b(
            "GET",
            "/me/top/tracks",
            params=_clear_nones(
                dict(limit=limit, offset=offset, time_range=time_range)
            ),
        )

    @returns(models.SavedAlbumsPaging)
    def saved_albums(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        market: Optional[str] = None,
    ) -> Request:
        return self.b(
            "GET",
            "/me/albums",
            params=_clear_nones(
                dict(market=market, limit=limit, offset=offset)
            ),
        )

    @returns(models.SavedTracksPaging)
    def saved_tracks(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        market: Optional[str] = None,
    ) -> Request:
        return self.b(
            "GET",
            "/me/tracks",
            params=_clear_nones(
                dict(limit=limit, offset=offset, market=market)
            ),
        )

    @returns(None)
    def remove_saved_albums(self, ids: Sequence[str]) -> Request:
        return self.b("DELETE", "/me/albums", params=dict(ids=",".join(ids)))

    @returns(None)
    def add_saved_albums(self, ids: Sequence[str]) -> Request:
        return self.b("PUT", "/me/albums", params=dict(ids=",".join(ids)))

    @returns(None)
    def remove_saved_tracks(self, ids: Sequence[str]) -> Request:
        return self.b("DELETE", "/me/tracks", params=dict(ids=",".join(ids)))

    @returns(None)
    def add_saved_tracks(self, ids: Sequence[str]) -> Request:
        return self.b("PUT", "/me/tracks", params=dict(ids=",".join(ids)))


class Follow(RequestBuilder):
    @returns(None)
    def follow_artists(self, ids: Sequence[str]) -> Request:
        return self.b(
            "PUT",
            "/me/following",
            params=dict(type="artist"),
            body=dict(ids=ids),
        )

    @returns(None)
    def follow_playlist(
        self, playlist_id: str, public: bool = None
    ) -> Request:
        return self.b(
            "PUT",
            "/playlists/{}/followers".format(playlist_id),
            body=dict(public=public) if public is not None else None,
        )

    @returns(None)
    def unfollow_playlist(self, playlist_id: str) -> Request:
        return self.b("DELETE", "/playlists/{}/followers".format(playlist_id))


class Browse(RequestBuilder):
    @returns(models.Recommendations)
    def recommendations(
        self,
        seed_artists: Optional[Sequence[str]] = None,
        seed_tracks: Optional[Sequence[str]] = None,
        seed_genres: Optional[Sequence[str]] = None,
        limit: Optional[int] = None,
        **audio_features
    ) -> Request:
        """
        :param seed_artists: List of seed artist ids
        :param seed_tracks: List of seed track ids
        :param seed_genres: List of seed genres
        :param limit: Max number of tracks
        :param \\**audio_features: min_*/max_*/target_* audio feature filters
        """
        params: Dict[str, Any] = {}
        if seed_artists:
            params["seed_artists"] = ",".join(seed_artists)
        if seed_tracks:
            params["seed_tracks"] = ",".join(seed_tracks)
        if seed_genres:
            params["seed_genres"] = ",".join(seed_genres)
        if limit is not None:
            params["limit"] = limit
        params.update(audio_features)
        return self.b("GET", "/recommendations", params=params)

    @returns(models.NewAlbumReleases)
    def new_releases(
        self,
        country: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Request:
        return self.b(
            "GET",
            "/browse/new-releases",
            params=_clear_nones(
                dict(country=country, limit=limit, offset=offset)
            ),
        )


class Auth(RequestBuilder):
    @returns(models.ClientCredentialsToken)
    def get_token_from_client_credentials(self) -> Request:
        """
        Get an authorization token from client credentials
        using :py:attr:`~client.client_id` and :py:attr:`~client.client_secret`
        """
        return self.b(
            "POST",
            self.client.token_url,
            params={"grant_type": "client_credentials"},
            headers={
                "Authorization": _basic_auth_str(  # type: ignore
                    self.client.client_id, self.client.client_secret
                ),
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )

    @returns(models.UserToken)
    def get_token_from_refresh_token(
        self, refresh_token: str = None
    ) -> Request:
        """
        Get a fresh authorization token from a refresh token

        :param refresh_token: A Spotify refresh token, if `None` will
            use the refresh token stored on client instance.
        """
        if refresh_token:
            refresh = refresh_token
        elif self.client.token and self.client.token.refresh_token:
            refresh = self.client.token.refresh_token
        else:
            raise ValueError(
                "No refresh token passed and no token on instance"
            )

        return self.b(
            "POST",
            self.client.token_url,
            params=dict(refresh_token=refresh, grant_type="refresh_token"),
            headers={
                "Authorization": _basic_auth_str(  # type: ignore
                    self.client.client_id, self.client.client_secret
                ),
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )

    @returns(models.UserToken)
    def get_token_from_code(self, response_code: str, **kwargs) -> Request:
        """
        Get an authorization token from an OAuth response code

        :param response_code: The response code from redirect URL
        :param \\**kwargs: Additional keyword arguments that
           override instance attributes and are sent to the
           token API as query params
        """
        params = dict(
            redirect_uri=self.client.redirect_uri,
            code=response_code,
            grant_type="authorization_code",
            scope=self.client.scope,
            state=self.client.state,
        )
        params.update(kwargs)

        return self.b(
            "POST",
            self.client.token_url,
            params=params,
            headers={
                "Authorization": _basic_auth_str(  # type: ignore
                    self.client.client_id, self.client.client_secret
                ),
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )


def _clear_nones(params: dict):
    return {key: value for key, value in params.items() if value is not None}


def _basic_auth_str(username: str, password: str, prefix="Basic ") -> str:
    """
    Create a base64 basic auth header value from username/password
    """
    return (
        prefix
        + b64encode(b":".join((username.encode(), password.encode()))).decode()
    )



from typing import TypeVar

from spoffy.client.base import AsyncClient, SyncClient
from spoffy.util import get_page_url

from spoffy.modules.modules import (
    Auth,
    AsyncAuth,
    Artists,
    AsyncArtists,
    Player,
    AsyncPlayer,
    Tracks,
    AsyncTracks,
    Albums,
    AsyncAlbums,
    Playlists,
    AsyncPlaylists,
    Search,
    AsyncSearch,
    Library,
    AsyncLibrary,
    Users,
    AsyncUsers,
    Follow,
    AsyncFollow,
    Browse,
    AsyncBrowse,
)


TPage = TypeVar("TPage")


class AsyncSpotify:
    """
    Async Spotify API wrapper. Wraps the client object
    and exposes the Spotify web API organized into modules.

    Use like so:

    >>> await AsyncSpotify(...).tracks.audio_features('sometrackid')

    :ivar ~AsyncSpotify.client: The underlying client instance
    :vartype ~AsyncSpotify.client: :class:`AsyncClient`
    :ivar ~AsyncSpotify.auth: Authorization methods
    :vartype ~AsyncSpotify.auth: :py:class:`~spoffy.modules.modules.AsyncAuth`
    :ivar ~AsyncSpotify.albums: Access to album endpoints
    :vartype ~AsyncSpotify.albums: :py:class:`~spoffy.modules.modules.AsyncAlbums`
    :ivar ~AsyncSpotify.tracks: Access to track endpoints
    :vartype ~AsyncSpotify.tracks: :py:class:`~spoffy.modules.modules.AsyncTracks`
    :ivar ~AsyncSpotify.playlists: Access to playlist endpoints
    :vartype ~AsyncSpotify.playlists: :py:class:`~spoffy.modules.modules.AsyncPlaylists`
    :ivar ~AsyncSpotify.artists: Access to artist endpoints
    :vartype ~AsyncSpotify.artists: :py:class:`~spoffy.modules.modules.AsyncArtists`
    :ivar ~AsyncSpotify.search: Access to search endpoints
    :vartype ~AsyncSpotify.search: :py:class:`~spoffy.modules.modules.AsyncSearch`
    :ivar ~AsyncSpotify.player: Access to player endpoints
    :vartype ~AsyncSpotify.player: :py:class:`~spoffy.modules.modules.AsyncPlayer`
    :ivar ~AsyncSpotify.library: Access to music library endpoints
    :vartype ~AsyncSpotify.library: :py:class:`~spoffy.modules.modules.AsyncLibrary`
    :ivar ~AsyncSpotify.users: Access to user endpoints
    :vartype ~AsyncSpotify.users: :py:class:`~spoffy.modules.modules.AsyncUsers`
    :ivar ~AsyncSpotify.follow: Follow artists, playlists and users
    :vartype ~AsyncSpotify.follow: :py:class:`~spoffy.modules.modules.AsyncFollow`
    :ivar ~AsyncSpotify.browse: Recommendations and new releases
    :vartype ~AsyncSpotify.browse: :py:class:`~spoffy.modules.modules.AsyncBrowse`


    :param client: An async client instance

    """  # noqa

    def __init__(self, client: AsyncClient) -> None:
        self.client = client

        self.artists = AsyncArtists(self.client)
        self.auth = AsyncAuth(self.client)
        self.player = AsyncPlayer(self.client)
        self.tracks = AsyncTracks(self.client)
        self.albums = AsyncAlbums(self.client)
        self.playlists = AsyncPlaylists(self.client)
        self.search = AsyncSearch(self.client)
        self.library = AsyncLibrary(self.client)
        self.users = AsyncUsers(self.client)
        self.follow = AsyncFollow(self.client)
        self.browse = AsyncBrowse(self.client)

    async def next_page(self, page: TPage) -> TPage:
        url = get_page_url(page, direction="next")
        return self.client.load(
            await self.client.request(self.client.prepare_request("GET", url)),
            page.__class__,
        )

    async def previous_page(self, page: TPage) -> TPage:
        url = get_page_url(page, direction="previous")
        return self.client.load(
            await self.client.request(self.client.prepare_request("GET", url)),
            page.__class__,
        )


class SyncSpotify:
    """
    Synchronous Spotify API wrapper. Wraps the client object
    and exposes the Spotify web API organized into modules.

    Use like so:

    >>> SyncSpotify(...).tracks.audio_features('sometrackid')

    :ivar ~SyncSpotify.client: The underlying client instance
    :vartype ~SyncSpotify.client: :py:class:`~SyncClient`
    :ivar ~SyncSpotify.auth: Authorization methods
    :vartype ~SyncSpotify.auth: :py:class:`~spoffy.modules.modules.Auth`
    :ivar ~SyncSpotify.albums: Access to album endpoints
    :vartype ~SyncSpotify.albums: :py:class:`~spoffy.modules.modules.Albums`
    :ivar ~SyncSpotify.tracks: Access to track endpoints
    :vartype ~SyncSpotify.tracks: :py:class:`~spoffy.modules.modules.Tracks`
    :ivar ~SyncSpotify.playlists: Access to playlist endpoints
    :vartype ~SyncSpotify.playlists: :py:class:`~spoffy.modules.modules.Playlists`
    :ivar ~SyncSpotify.artists: Access to artist endpoints
    :vartype ~SyncSpotify.artists: :py:class:`~spoffy.modules.modules.Artists`
    :ivar ~SyncSpotify.search: Access to search endpoints
    :vartype ~SyncSpotify.search: :py:class:`~spoffy.modules.modules.Search`
    :ivar ~SyncSpotify.player: Access to player endpoints
    :vartype ~SyncSpotify.player: :py:class:`~spoffy.modules.modules.Player`
    :ivar ~SyncSpotify.library: Access to music library endpoints
    :vartype ~SyncSpotify.library: :py:class:`~spoffy.modules.modules.Library`
    :ivar ~SyncSpotify.users: Access to user endpoints
    :vartype ~SyncSpotify.users: :py:class:`~spoffy.modules.modules.Users`
    :ivar ~SyncSpotify.follow: Follow artists, playlists and users
    :vartype ~SyncSpotify.follow: :py:class:`~spoffy.modules.modules.Follow`
    :ivar ~Spotify.browse: Recommendations and new releases
    :vartype ~Spotify.browse: :py:class:`~spoffy.modules.modules.Browse`

    :param client: An sync client instance

    """  # noqa

    def __init__(self, client: SyncClient) -> None:
        self.client = client

        self.artists = Artists(self.client)
        self.auth = Auth(self.client)
        self.player = Player(self.client)
        self.tracks = Tracks(self.client)
        self.albums = Albums(self.client)
        self.playlists = Playlists(self.client)
        self.search = Search(self.client)
        self.library = Library(self.client)
        self.users = Users(self.client)
        self.follow = Follow(self.client)
        self.browse = Browse(self.client)

    def next_page(self, page: TPage) -> TPage:
        url = get_page_url(page, direction="next")
        return self.client.load(
            self.client.request(self.client.prepare_request("GET", url)),
            page.__class__,
        )

    def previous_page(self, page: TPage) -> TPage:
        url = get_page_url(page, direction="previous")
        return self.client.load(
            self.client.request(self.client.prepare_request("GET", url)),
            page.__class__,
        )



from typing import Optional as Opt, List

from spoffy.models.core import Track, ExternalUrls
from spoffy.models.paging import CursorPaging
from spoffy.models.base import SpotifyObject


class Device(SpotifyObject):
    id: Opt[str]
    is_active: bool
    is_private_session: bool
    is_restricted: bool
    name: str
    type: str
    volume_percent: Opt[int]


class DevicesCollection(SpotifyObject):
    devices: List[Device]


class Context(SpotifyObject):
    uri: str
    href: Opt[str]
    external_urls: Opt[ExternalUrls]
    type: str


class CurrentPlayback(SpotifyObject):
    device: Device
    repeat_state: str
    shuffle_state: bool
    context: Opt[Context]
    timestamp: int
    progress_ms: Opt[int]
    is_playing: bool
    item: Opt[Track]
    currently_playing_type: str


class PlayHistoryItem(SpotifyObject):
    track: Track
    played_at: str
    context: Opt[Context]


class PlayHistoryPaging(CursorPaging):
    items: List[PlayHistoryItem]



from spoffy.spotify import SyncSpotify, AsyncSpotify
from spoffy.client.base import SyncClient, AsyncClient
from spoffy.client.common import ClientCommon
from spoffy.exceptions import (
    SpotifyException,
    SpotifyUnauthorized,
    SpotifyPremiumRequired,
)

__version__ = "0.4.0"


__all__ = [
    "SyncSpotify",
    "AsyncSpotify",
    "SyncClient",
    "AsyncClient",
    "ClientCommon",
    "SpotifyException",
    "SpotifyUnauthorized",
    "SpotifyPremiumRequired",
]



from typing import List, Union

from spoffy.models.core import Track, Artist
from spoffy.models.audiofeatures import AudioFeatures
from spoffy.models.base import SpotifyObject


class TracksCollection(SpotifyObject):
    tracks: List[Union[Track, None]]


class AudioFeaturesCollection(SpotifyObject):
    audio_features: List[Union[AudioFeatures, None]]


class ArtistsCollection(SpotifyObject):
    artists: List[Union[Artist, None]]


class RelatedArtistsCollection(SpotifyObject):
    artists: List[Artist]



from typing import List

from spoffy.models.paging import OffsetPaging
from spoffy.models.core import Artist, Track, AlbumSimple
from spoffy.models.playlists import PlaylistSimple
from spoffy.models.base import SpotifyObject


class SearchArtistsPaging(OffsetPaging):
    items: List[Artist]


class SearchTracksPaging(OffsetPaging):
    items: List[Track]


class SearchAlbumsPaging(OffsetPaging):
    items: List[AlbumSimple]


class SearchPlaylistsPaging(OffsetPaging):
    items: List[PlaylistSimple]


class SearchResults(SpotifyObject):
    artists: SearchArtistsPaging
    tracks: SearchTracksPaging
    albums: SearchAlbumsPaging
    playlists: SearchPlaylistsPaging



from spoffy.models.paging import Paging
from spoffy.models.image import Image
from spoffy.models.playlists import (
    Playlist,
    PlaylistSimple,
    PlaylistTracksPaging,
    PlaylistOwner,
    PlaylistSnapshotId,
    TracksHref,
    PlaylistSimplePaging,
)
from spoffy.models.core import (
    Artist,
    Track,
    Album,
    AlbumTracksPaging,
    AlbumSimplePaging,
    NewAlbumReleases,
    ExternalUrls,
    TrackSimple,
)
from spoffy.models.collections import (
    TracksCollection,
    AudioFeaturesCollection,
    ArtistsCollection,
    RelatedArtistsCollection,
)
from spoffy.models.audiofeatures import AudioFeatures
from spoffy.models.search import SearchResults
from spoffy.models.personalization import TopTracksPaging, TopArtistsPaging
from spoffy.models.library import SavedAlbumsPaging, SavedTracksPaging
from spoffy.models.player import (
    Device,
    CurrentPlayback,
    DevicesCollection,
    PlayHistoryPaging,
)
from spoffy.models.token import ClientCredentialsToken, UserToken, Token
from spoffy.models.users import UserPublic, UserPrivate
from spoffy.models.recommendations import RecommendationSeed, Recommendations


__all__ = [
    "Playlist",
    "PlaylistTracksPaging",
    "PlaylistSimple",
    "Artist",
    "Track",
    "Album",
    "AlbumTracksPaging",
    "TracksCollection",
    "AudioFeaturesCollection",
    "AudioFeatures",
    "SearchResults",
    "TopTracksPaging",
    "TopArtistsPaging",
    "SavedAlbumsPaging",
    "SavedTracksPaging",
    "Device",
    "CurrentPlayback",
    "DevicesCollection",
    "PlayHistoryPaging",
    "UserToken",
    "ClientCredentialsToken",
    "UserPublic",
    "UserPrivate",
    "Token",
    "Paging",
    "Image",
    "ExternalUrls",
    "PlaylistOwner",
    "TracksHref",
    "PlaylistSnapshotId",
    "TrackSimple",
    "RecommendationSeed",
    "Recommendations",
    "AlbumSimplePaging",
    "NewAlbumReleases",
    "ArtistsCollection",
    "RelatedArtistsCollection",
    "PlaylistSimplePaging",
]



def get_page_url(page, direction="next") -> str:
    target_keys = ["artists", "playlists", "albums", "tracks"]
    page_url: str = getattr(page, direction, None)
    if page_url:
        return page_url

    for key in target_keys:
        nested = getattr(page, key, None)
        if nested:
            page_url = getattr(nested, direction)
    if page_url:
        return page_url
    raise AttributeError(
        "{} url not found on object of type {}".format(direction, type(page))
    )



from typing import List

from spoffy.models.core import TrackSimple
from spoffy.models.base import SpotifyObject


class RecommendationSeed(SpotifyObject):
    initialPoolSize: int
    afterFilteringSize: int
    afterRelinkingSize: int
    id: str
    type: str
    href: str


class Recommendations(SpotifyObject):
    tracks: List[TrackSimple]
    seeds: List[RecommendationSeed]
